/*
 * Copyright (c) 2006 Zmanda Inc.
 *
 * Derived from client-src/sendbackup-gnutar.c:
 *
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991-1999 University of Maryland at College Park
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: the Amanda Development Team.  Its members are listed in a
 * file named AUTHORS, in the root directory of this distribution.
 */
/* 
 * $Id: sendbackup-star.c 9003 2007-10-26 19:39:58Z jfranks $
 *
 * send backup data using star
 */

#include "amanda.h"
#include "sendbackup.h"
#include "amandates.h"
#include "clock.h"
#include "util.h"
#include "getfsent.h"			/* for amname_to_dirname lookup */
#include "version.h"
#include "conffile.h"

static amregex_t re_table[] = {
  /* star prints the size in 10k blocks */
  AM_SIZE_RE("star: [0-9][0-9]* blocks", 10240, 1),
  AM_NORMAL_RE("^Type of this level "),
  AM_NORMAL_RE("^Date of this level "),
  AM_NORMAL_RE("^Date of last level "),
  AM_NORMAL_RE("^Dump record  level "),
  AM_NORMAL_RE("^Throughput"),
  AM_NORMAL_RE("^.*is sparse$"),

  /*
   * STAR produces a few error messages when files are modified or
   * removed while it is running.  They may cause data to be lost, but
   * then they may not.  We shouldn't consider them NORMAL until
   * further investigation.
   */
#ifdef IGNORE_TAR_ERRORS
  AM_NORMAL_RE("^.*shrunk*$"),
  AM_NORMAL_RE("^.*changed size.*$"),
  AM_NORMAL_RE("^.*Cannot listxattr for.*$"),
  AM_NORMAL_RE("^.Cannot: stat .*$"),
  AM_NORMAL_RE("^.Missing links .*$"),
  AM_NORMAL_RE("^.Cannot get xattr.*$"),
  AM_NORMAL_RE("^.Cannot.*acl.*$"),
#endif
  
  AM_NORMAL_RE("^star: dumped [0-9][0-9]* (tar )?files"),
  AM_NORMAL_RE("^.*The following problems occurred during .* processing.*$"),
  AM_NORMAL_RE("^.*Processed all possible files, despite earlier errors.*$"),
  AM_NORMAL_RE("^.*not written due to problems during backup.*$"),

  AM_STRANGE_RE("^Perform a level 0 dump first.*$"),

  /* catch-all: DMP_STRANGE is returned for all other lines */
  AM_STRANGE_RE(NULL)
};

extern char *efile;

int cur_level;
char *cur_disk;
time_t cur_dumptime;

static char *amandates_file;
static void start_backup(char *host, char *disk, char *amdevice, int level,
		char *dumpdate, int dataf, int mesgf, int indexf);
static void end_backup(int status);

/*
 *  doing similar to $ star | compression | encryption 
 */
static void
start_backup(
    char *	host,
    char *	disk,
    char *	amdevice,
    int		level,
    char *	dumpdate,
    int		dataf,
    int		mesgf,
    int		indexf)
{
    int dumpin, dumpout, compout;
    char *cmd = NULL;
    char *indexcmd = NULL;
    char *dirname = NULL;
    char *fsname = NULL;
    int l;
    char dumptimestr[80];
    struct tm *gmtm;
    amandates_t *amdates;
    time_t prev_dumptime;
    char *error_pn = NULL;
    char *compopt  = NULL;
    char *encryptopt = skip_argument;
    char *qdisk;
    FILE *exclude;
    char *aexc = NULL;
    sle_t *excl;
    char *quoted;

    (void)dumpdate;	/* Quiet unused parameter warning */

    error_pn = stralloc(get_pname());

    qdisk = quote_string(disk);
    dbprintf(_("start: %s:%s lev %d\n"), host, qdisk, level);

    g_fprintf(stderr, _("%s: start [%s:%s level %d]\n"),
	    get_pname(), host, qdisk, level);

    /*  apply client-side encryption here */
    if ( options->encrypt == ENCRYPT_CUST ) {
	encpid = pipespawn(options->clnt_encrypt, STDIN_PIPE,
			&compout, &dataf, &mesgf, 
			options->clnt_encrypt, encryptopt, NULL);
	dbprintf(_("pid %ld: %s\n"), (long)encpid, options->clnt_encrypt);
    } else {
	compout = dataf;
	encpid = -1;
    } 
     /*  now do the client-side compression */
    if(options->compress == COMP_FAST || options->compress == COMP_BEST) {
          compopt = skip_argument;
#if defined(COMPRESS_BEST_OPT) && defined(COMPRESS_FAST_OPT)
	if(options->compress == COMP_BEST) {
	    compopt = COMPRESS_BEST_OPT;
	} else {
	    compopt = COMPRESS_FAST_OPT;
	}
#endif
	comppid = pipespawn(COMPRESS_PATH, STDIN_PIPE,
			    &dumpout, &compout, &mesgf,
			    COMPRESS_PATH, compopt, NULL);
	if(compopt != skip_argument) {
	    dbprintf(_("pid %ld: %s %s\n"), (long)comppid, COMPRESS_PATH, compopt);
	} else {
	    dbprintf(_("pid %ld: %s\n"), (long)comppid, COMPRESS_PATH);
	}
     } else if (options->compress == COMP_CUST) {
        compopt = skip_argument;
	comppid = pipespawn(options->clntcompprog, STDIN_PIPE,
			    &dumpout, &compout, &mesgf,
			    options->clntcompprog, compopt, NULL);
	dbprintf(_("star-cust: pid %ld: %s"),
		(long)comppid, options->clntcompprog);
	if(compopt != skip_argument) {
	    dbprintf(_("star-cust: pid %ld: %s %s\n"),
		(long)comppid, options->clntcompprog, compopt);
	} else {
	    dbprintf(_("star-cust: pid %ld: %s\n"),
		(long)comppid, options->clntcompprog);
	}
    } else {
	dumpout = compout;
	comppid = -1;
    }

    /* find previous dump time */

    amandates_file = getconf_str(CNF_AMANDATES);
    if(!start_amandates(amandates_file, 0)) {
	error(_("error [opening %s: %s]"), amandates_file, strerror(errno));
	/*NOTREACHED*/
    }

    amdates = amandates_lookup(disk);

    prev_dumptime = EPOCH;
    for(l = 0; l < level; l++) {
	if(amdates->dates[l] > prev_dumptime)
	    prev_dumptime = amdates->dates[l];
    }

    finish_amandates();
    free_amandates();

    gmtm = gmtime(&prev_dumptime);
    g_snprintf(dumptimestr, SIZEOF(dumptimestr),
		"%04d-%02d-%02d %2d:%02d:%02d GMT",
		gmtm->tm_year + 1900, gmtm->tm_mon+1, gmtm->tm_mday,
		gmtm->tm_hour, gmtm->tm_min, gmtm->tm_sec);

    dbprintf(_("star: doing level %d dump from date: %s\n"), level, dumptimestr);

    dirname = amname_to_dirname(amdevice);
    cur_dumptime = time(0);
    cur_level = level;
    cur_disk = stralloc(disk);
#ifdef STAR
#  define PROGRAM_STAR STAR
#else
#  define PROGRAM_STAR "star"
#endif
    indexcmd = vstralloc(
			 PROGRAM_STAR,
			 " -t -f -",
			 " -xattr -acl H=exustar",
			 " errctl='WARN|SAMEFILE|DIFF|GROW|SHRINK|SPECIALFILE|GETXATTR|BADACL *'",
			 " 2>/dev/null",
			 " | sed -e \'s%^\\(.*\\) -> .*%\\1%\'",
			 " | sed -e \'s%^[^/]%/&%\'",
			 NULL);
    {
	int nb_exclude = 0;
	int nb_include = 0;
	char **my_argv;
	int my_argc = 0;
	char *file_include = NULL;
	char levelstr[NUM_STR_SIZE + 7];
	char *s;

	if(options->exclude_file) nb_exclude+=options->exclude_file->nb_element;
	if(options->exclude_list) nb_exclude+=options->exclude_list->nb_element;
	if(options->include_file) nb_include+=options->include_file->nb_element;
	if(options->include_list) nb_include+=options->include_list->nb_element;

	if(nb_include > 0) file_include = build_include(disk, amdevice, options, 0);

	my_argv = alloc(SIZEOF(char *) * (32 + (nb_exclude*2)+(nb_include*2)));

	cmd = vstralloc(libexecdir, "/", "runstar", versionsuffix(), NULL);
	info_tapeheader();

	start_index(options->createindex, dumpout, mesgf, indexf, indexcmd);

	fsname = vstralloc("fs-name=", dirname, NULL);
	for (s = fsname; *s != '\0'; s++) {
	    if (iscntrl((int)*s))
		*s = '-';
	}
	dbprintf("star: %s: %s\n", cmd, fsname);

	strcpy(levelstr, "-level=");
	g_snprintf(levelstr+7, SIZEOF(levelstr)-7, "%d", level);

 	my_argv[my_argc++] = stralloc("runstar");
	if (g_options->config)
	    my_argv[my_argc++] = stralloc(g_options->config);
	else
	    my_argv[my_argc++] = stralloc("NOCONFIG");

	my_argv[my_argc++] = stralloc("star");
	my_argv[my_argc++] = stralloc("-c");
	my_argv[my_argc++] = stralloc("-f");
	my_argv[my_argc++] = stralloc("-");
	
	my_argv[my_argc++] = stralloc("-C");
#if defined(__CYGWIN__)
	{
	    char tmppath[PATH_MAX];

	    cygwin_conv_to_full_posix_path(dirname, tmppath);
	    my_argv[my_argc++] = stralloc(tmppath);
	}
#else
	my_argv[my_argc++] = stralloc(dirname);
#endif
	my_argv[my_argc++] = stralloc(fsname);
	my_argv[my_argc++] = stralloc("-xdev");

	my_argv[my_argc++] = stralloc("-link-dirs");
	my_argv[my_argc++] = stralloc(levelstr);
	my_argv[my_argc++] = stralloc("-wtardumps");
	my_argv[my_argc++] = stralloc("-xattr");
	my_argv[my_argc++] = stralloc("-acl");
	my_argv[my_argc++] = stralloc("H=exustar");
	my_argv[my_argc++] = stralloc("errctl=WARN|SAMEFILE|DIFF|GROW|SHRINK|SPECIALFILE|GETXATTR|BADACL *");
	my_argv[my_argc++] = stralloc("-sparse");
	my_argv[my_argc++] = stralloc("-dodesc");

	if(nb_exclude != 0) {
	    my_argv[my_argc++] = stralloc("-not");
	    if(options->exclude_file) {
		for(excl = options->exclude_file->first; excl != NULL;
		    excl = excl->next) {
			my_argv[my_argc++] = vstralloc("pat=", excl->name, NULL);
		}
	    }

	    if(options->exclude_list) {
		for(excl = options->exclude_list->first; excl != NULL;
		    excl = excl->next) {
		    char *exclname = fixup_relative(excl->name, amdevice);
		    if((exclude = fopen(exclname, "r")) != NULL) {
			while ((aexc = agets(exclude)) != NULL) {
			    if (aexc[0] == '\0') {
				amfree(aexc);
				continue;
			    }
			    my_argv[my_argc++] = vstralloc("pat=", aexc, NULL);
			    amfree(aexc);
			}
			fclose(exclude);
		    }
		    else {
			quoted = quote_string(exclname);
			dbprintf(_("Can't open exclude file %s (%s)\n"),
				  quoted, strerror(errno));
			amfree(quoted);
		    }
		    amfree(exclname);
		}
	    }
	}

	if (file_include) {
	    my_argv[my_argc++] = vstralloc("list=", file_include, NULL);
	} else {
	    my_argv[my_argc++] = vstralloc(".", NULL);
	}
	my_argv[my_argc] = NULL;

	dumppid = pipespawnv(cmd, STDIN_PIPE,
			     &dumpin, &dumpout, &mesgf, my_argv);
	tarpid = dumppid;

        amfree(fsname);
	amfree(file_include);
	while (my_argc-- > 0)
	    amfree(my_argv[my_argc]);
	amfree(my_argv);
    }
    dbprintf(_("%s: pid %ld\n"), cmd, (long)dumppid);

    amfree(dirname);
    amfree(cmd);
    amfree(indexcmd);
    amfree(qdisk);
    amfree(error_pn);

    /* close the write ends of the pipes */
    aclose(dumpin);
    aclose(dumpout);
    aclose(compout);
    aclose(dataf);
    aclose(mesgf);
    if (options->createindex)
	aclose(indexf);

}

static void
end_backup(
    int		goterror)
{
    if(!options->no_record && !goterror) {
        if(!start_amandates(amandates_file, 1)) {
	    g_fprintf(stderr, _("%s: warning [opening %s: %s]"), get_pname(),
		    amandates_file, strerror(errno));
	}
	else {
	    amandates_updateone(cur_disk, cur_level, cur_dumptime);
	    finish_amandates();
	    free_amandates();
	}
    }
}

backup_program_t star_program = {
  "STAR",
#ifdef STAR
  STAR, STAR,
#else
  "star", "star",
#endif
  re_table, start_backup, end_backup
};
